@@ -22,17 +22,15 @@ See :doc:`/central_scheduler` for more info.
 """
 
 import collections
-import inspect
+from collections.abc import MutableSet
 import json
 
 from luigi.batch_notifier import BatchNotifier
 
-try:
-    import cPickle as pickle
-except ImportError:
-    import pickle
+import pickle
 import functools
 import hashlib
+import inspect
 import itertools
 import logging
 import os
@@ -40,8 +38,6 @@ import re
 import time
 import uuid
 
-from luigi import six
-
 from luigi import configuration
 from luigi import notifications
 from luigi import parameter
@@ -51,6 +47,8 @@ from luigi.task_status import DISABLED, DONE, FAILED, PENDING, RUNNING, SUSPENDE
 from luigi.task import Config
 from luigi.parameter import ParameterVisibility
 
+from luigi.metrics import MetricsCollectors
+
 logger = logging.getLogger(__name__)
 
 UPSTREAM_RUNNING = 'UPSTREAM_RUNNING'
@@ -97,12 +95,11 @@ def rpc_method(**request_args):
     def _rpc_method(fn):
         # If request args are passed, return this function again for use as
         # the decorator function with the request args attached.
-        fn_args = inspect.getargspec(fn)
-
-        assert not fn_args.varargs
-        assert fn_args.args[0] == 'self'
-        all_args = fn_args.args[1:]
-        defaults = dict(zip(reversed(all_args), reversed(fn_args.defaults or ())))
+        args, varargs, varkw, defaults, kwonlyargs, kwonlydefaults, ann = inspect.getfullargspec(fn)
+        assert not varargs
+        first_arg, *all_args = args
+        assert first_arg == 'self'
+        defaults = dict(zip(reversed(all_args), reversed(defaults or ())))
         required_args = frozenset(arg for arg in all_args if arg not in defaults)
         fn_name = fn.__name__
 
@@ -147,57 +144,19 @@ class scheduler(Config):
 
     send_messages = parameter.BoolParameter(default=True)
 
+    metrics_collector = parameter.EnumParameter(enum=MetricsCollectors, default=MetricsCollectors.default)
+    metrics_custom_import = parameter.OptionalStrParameter(default=None)
+
+    stable_done_cooldown_secs = parameter.IntParameter(default=10,
+                                                       description="Sets cooldown period to avoid running the same task twice")
+    """
+    Sets a cooldown period in seconds after a task was completed, during this period the same task will not accepted by the scheduler.
+    """
+
     def _get_retry_policy(self):
         return RetryPolicy(self.retry_count, self.disable_hard_timeout, self.disable_window)
 
 
-class Failures(object):
-    """
-    This class tracks the number of failures in a given time window.
-
-    Failures added are marked with the current timestamp, and this class counts
-    the number of failures in a sliding time window ending at the present.
-    """
-
-    def __init__(self, window):
-        """
-        Initialize with the given window.
-
-        :param window: how long to track failures for, as a float (number of seconds).
-        """
-        self.window = window
-        self.failures = collections.deque()
-        self.first_failure_time = None
-
-    def add_failure(self):
-        """
-        Add a failure event with the current timestamp.
-        """
-        failure_time = time.time()
-
-        if not self.first_failure_time:
-            self.first_failure_time = failure_time
-
-        self.failures.append(failure_time)
-
-    def num_failures(self):
-        """
-        Return the number of failures in the window.
-        """
-        min_time = time.time() - self.window
-
-        while self.failures and self.failures[0] < min_time:
-            self.failures.popleft()
-
-        return len(self.failures)
-
-    def clear(self):
-        """
-        Clear the failure queue.
-        """
-        self.failures.clear()
-
-
 def _get_default(x, default):
     if x is not None:
         return x
@@ -205,7 +164,7 @@ def _get_default(x, default):
         return default
 
 
-class OrderedSet(collections.MutableSet):
+class OrderedSet(MutableSet):
     """
     Standard Python OrderedSet recipe found at http://code.activestate.com/recipes/576694/
 
@@ -273,7 +232,7 @@ class OrderedSet(collections.MutableSet):
         return set(self) == set(other)
 
 
-class Task(object):
+class Task:
     def __init__(self, task_id, status, deps, resources=None, priority=0, family='', module=None,
                  params=None, param_visibilities=None, accepts_messages=False, tracking_url=None, status_message=None,
                  progress_percentage=None, retry_policy='notoptional'):
@@ -303,7 +262,8 @@ class Task(object):
         self.set_params(params)
         self.accepts_messages = accepts_messages
         self.retry_policy = retry_policy
-        self.failures = Failures(self.retry_policy.disable_window)
+        self.failures = collections.deque()
+        self.first_failure_time = None
         self.tracking_url = tracking_url
         self.status_message = status_message
         self.progress_percentage = progress_percentage
@@ -332,27 +292,53 @@ class Task(object):
             return False
 
     def add_failure(self):
-        self.failures.add_failure()
+        """
+        Add a failure event with the current timestamp.
+        """
+        failure_time = time.time()
+
+        if not self.first_failure_time:
+            self.first_failure_time = failure_time
+
+        self.failures.append(failure_time)
+
+    def num_failures(self):
+        """
+        Return the number of failures in the window.
+        """
+        min_time = time.time() - self.retry_policy.disable_window
+
+        while self.failures and self.failures[0] < min_time:
+            self.failures.popleft()
+
+        return len(self.failures)
 
     def has_excessive_failures(self):
-        if self.failures.first_failure_time is not None:
-            if (time.time() >= self.failures.first_failure_time + self.retry_policy.disable_hard_timeout):
+        if self.first_failure_time is not None:
+            if time.time() >= self.first_failure_time + self.retry_policy.disable_hard_timeout:
                 return True
 
-        logger.debug('%s task num failures is %s and limit is %s', self.id, self.failures.num_failures(), self.retry_policy.retry_count)
-        if self.failures.num_failures() >= self.retry_policy.retry_count:
+        logger.debug('%s task num failures is %s and limit is %s', self.id, self.num_failures(), self.retry_policy.retry_count)
+        if self.num_failures() >= self.retry_policy.retry_count:
             logger.debug('%s task num failures limit(%s) is exceeded', self.id, self.retry_policy.retry_count)
             return True
 
         return False
 
+    def clear_failures(self):
+        """
+        Clear the failures history
+        """
+        self.failures.clear()
+        self.first_failure_time = None
+
     @property
     def pretty_id(self):
         param_str = ', '.join(u'{}={}'.format(key, value) for key, value in sorted(self.public_params.items()))
         return u'{}({})'.format(self.family, param_str)
 
 
-class Worker(object):
+class Worker:
     """
     Structure for tracking worker activity and keeping their references.
     """
@@ -387,9 +373,9 @@ class Worker(object):
         num_self_tasks = len(self.tasks)
         num_state_tasks = sum(len(state._status_tasks[status]) for status in statuses)
         if num_self_tasks < num_state_tasks:
-            return six.moves.filter(lambda task: task.status in statuses, self.tasks)
+            return filter(lambda task: task.status in statuses, self.tasks)
         else:
-            return six.moves.filter(lambda task: self.id in task.workers, state.get_active_tasks_by_status(*statuses))
+            return filter(lambda task: self.id in task.workers, state.get_active_tasks_by_status(*statuses))
 
     def is_trivial_worker(self, state):
         """
@@ -430,9 +416,9 @@ class Worker(object):
         return self.id
 
 
-class SimpleTaskState(object):
+class SimpleTaskState:
     """
-    Keep track of the current state and handle persistance.
+    Keep track of the current state and handle persistence.
 
     The point of this class is to enable other ways to keep state, eg. by using a database
     These will be implemented by creating an abstract base class that this and other classes
@@ -445,6 +431,7 @@ class SimpleTaskState(object):
         self._status_tasks = collections.defaultdict(dict)
         self._active_workers = {}  # map from id to a Worker object
         self._task_batchers = {}
+        self._metrics_collector = None
 
     def get_state(self):
         return self._tasks, self._active_workers, self._task_batchers
@@ -476,16 +463,16 @@ class SimpleTaskState(object):
 
             self.set_state(state)
             self._status_tasks = collections.defaultdict(dict)
-            for task in six.itervalues(self._tasks):
+            for task in self._tasks.values():
                 self._status_tasks[task.status][task.id] = task
         else:
             logger.info("No prior state file exists at %s. Starting with empty state", self._state_path)
 
     def get_active_tasks(self):
-        return six.itervalues(self._tasks)
+        return self._tasks.values()
 
     def get_active_tasks_by_status(self, *statuses):
-        return itertools.chain.from_iterable(six.itervalues(self._status_tasks[status]) for status in statuses)
+        return itertools.chain.from_iterable(self._status_tasks[status].values() for status in statuses)
 
     def get_active_task_count_for_status(self, status):
         if status:
@@ -526,10 +513,10 @@ class SimpleTaskState(object):
 
     def re_enable(self, task, config=None):
         task.scheduler_disable_time = None
-        task.failures.clear()
+        task.clear_failures()
         if config:
             self.set_status(task, FAILED, config)
-            task.failures.clear()
+            task.clear_failures()
 
     def set_batch_running(self, task, batch_id, worker_id):
         self.set_status(task, BATCH_RUNNING)
@@ -573,7 +560,7 @@ class SimpleTaskState(object):
                         'disabled for {persist} seconds'.format(
                             failures=task.retry_policy.retry_count,
                             task=task.id,
-                            window=config.disable_window,
+                            window=task.retry_policy.disable_window,
                             persist=config.disable_persist,
                         ))
         elif new_status == DISABLED:
@@ -584,6 +571,7 @@ class SimpleTaskState(object):
             self._status_tasks[new_status][task.id] = task
             task.status = new_status
             task.updated = time.time()
+            self.update_metrics(task, config)
 
         if new_status == FAILED:
             task.retry = time.time() + config.retry_delay
@@ -630,7 +618,7 @@ class SimpleTaskState(object):
             self._status_tasks[task_obj.status].pop(task)
 
     def get_active_workers(self, last_active_lt=None, last_get_work_gt=None):
-        for worker in six.itervalues(self._active_workers):
+        for worker in self._active_workers.values():
             if last_active_lt is not None and worker.last_active >= last_active_lt:
                 continue
             last_get_work = worker.last_get_work
@@ -667,8 +655,16 @@ class SimpleTaskState(object):
             worker.disabled = True
             worker.tasks.clear()
 
+    def update_metrics(self, task, config):
+        if task.status == DISABLED:
+            self._metrics_collector.handle_task_disabled(task, config)
+        elif task.status == DONE:
+            self._metrics_collector.handle_task_done(task)
+        elif task.status == FAILED:
+            self._metrics_collector.handle_task_failed(task)
 
-class Scheduler(object):
+
+class Scheduler:
     """
     Async scheduler that can handle multiple workers, etc.
 
@@ -700,6 +696,8 @@ class Scheduler(object):
         if self._config.batch_emails:
             self._email_batcher = BatchNotifier()
 
+        self._state._metrics_collector = MetricsCollectors.get(self._config.metrics_collector, self._config.metrics_custom_import)
+
     def load(self):
         self._state.load()
 
@@ -780,6 +778,19 @@ class Scheduler(object):
             self._state.set_status(task, status, self._config)
         return {"task_id": task_id, "status": task.status}
 
+    @rpc_method()
+    def mark_as_done(self, task_id=None):
+        status = DONE
+        task = self._state.get_task(task_id)
+        if task is None:
+            return {"task_id": task_id, "status": None}
+
+        # we can force mark DONE for running or failed tasks
+        if task.status in {RUNNING, FAILED, DISABLED}:
+            self._update_task_history(task, status)
+            self._state.set_status(task, status, self._config)
+        return {"task_id": task_id, "status": task.status}
+
     @rpc_method()
     def add_task(self, task_id=None, status=PENDING, runnable=True,
                  deps=None, new_deps=None, expl=None, resources=None,
@@ -817,12 +828,16 @@ class Scheduler(object):
         if task is None or (task.status != RUNNING and not worker.enabled):
             return
 
+        # Ignore claims that the task is PENDING if it very recently was marked as DONE.
+        if status == PENDING and task.status == DONE and (time.time() - task.updated) < self._config.stable_done_cooldown_secs:
+            return
+
         # for setting priority, we'll sometimes create tasks with unset family and params
         if not task.family:
             task.family = family
         if not getattr(task, 'module', None):
             task.module = module
-        if not task.param_visibilities:
+        if not getattr(task, 'param_visibilities', None):
             task.param_visibilities = _get_default(param_visibilities, {})
         if not task.params:
             task.set_params(params)
@@ -875,7 +890,7 @@ class Scheduler(object):
             if batched_params:
                 unbatched_params = {
                     param: value
-                    for param, value in six.iteritems(task.params)
+                    for param, value in task.params.items()
                     if param not in batched_params
                 }
             else:
@@ -929,7 +944,7 @@ class Scheduler(object):
         if batched_params:
             unbatched_params = {
                 param: value
-                for param, value in six.iteritems(params)
+                for param, value in params.items()
                 if param not in batched_params
             }
         else:
@@ -973,6 +988,10 @@ class Scheduler(object):
             response = task.scheduler_message_responses.pop(message_id, None)
         return {"response": response}
 
+    @rpc_method()
+    def has_task_history(self):
+        return self._config.record_task_history
+
     @rpc_method()
     def is_pause_enabled(self):
         return {'enabled': self._config.pause_enabled}
@@ -1006,7 +1025,7 @@ class Scheduler(object):
 
     def _generate_retry_policy(self, task_retry_policy_dict):
         retry_policy_dict = self._config._get_retry_policy()._asdict()
-        retry_policy_dict.update({k: v for k, v in six.iteritems(task_retry_policy_dict) if v is not None})
+        retry_policy_dict.update({k: v for k, v in task_retry_policy_dict.items() if v is not None})
         return RetryPolicy(**retry_policy_dict)
 
     def _has_resources(self, needed_resources, used_resources):
@@ -1014,7 +1033,7 @@ class Scheduler(object):
             return True
 
         available_resources = self._resources or {}
-        for resource, amount in six.iteritems(needed_resources):
+        for resource, amount in needed_resources.items():
             if amount + used_resources[resource] > available_resources.get(resource, 1):
                 return False
         return True
@@ -1025,7 +1044,7 @@ class Scheduler(object):
             for task in self._state.get_active_tasks_by_status(RUNNING):
                 resources_running = getattr(task, "resources_running", task.resources)
                 if resources_running:
-                    for resource, amount in six.iteritems(resources_running):
+                    for resource, amount in resources_running.items():
                         used_resources[resource] += amount
         return used_resources
 
@@ -1174,7 +1193,7 @@ class Scheduler(object):
 
             if task.status == RUNNING and (task.worker_running in greedy_workers):
                 greedy_workers[task.worker_running] -= 1
-                for resource, amount in six.iteritems((getattr(task, 'resources_running', task.resources) or {})):
+                for resource, amount in (getattr(task, 'resources_running', task.resources) or {}).items():
                     greedy_resources[resource] += amount
 
             if self._schedulable(task) and self._has_resources(task.resources, greedy_resources):
@@ -1203,7 +1222,7 @@ class Scheduler(object):
                             greedy_workers[task_worker] -= 1
 
                             # keep track of the resources used in greedy scheduling
-                            for resource, amount in six.iteritems((task.resources or {})):
+                            for resource, amount in (task.resources or {}).items():
                                 greedy_resources[resource] += amount
 
                             break
@@ -1227,6 +1246,7 @@ class Scheduler(object):
             reply['batch_task_ids'] = [task.id for task in batched_tasks]
 
         elif best_task:
+            self.update_metrics_task_started(best_task)
             self._state.set_status(best_task, RUNNING, self._config)
             best_task.worker_running = worker_id
             best_task.resources_running = best_task.resources.copy()
@@ -1422,7 +1442,7 @@ class Scheduler(object):
             terms = search.split()
 
             def filter_func(t):
-                return all(term in t.pretty_id for term in terms)
+                return all(term.casefold() in t.pretty_id.casefold() for term in terms)
 
         tasks = self._state.get_active_tasks_by_status(status) if status else self._state.get_active_tasks()
         for task in filter(filter_func, tasks):
@@ -1487,12 +1507,12 @@ class Scheduler(object):
                 name=resource,
                 num_total=r_dict['total'],
                 num_used=r_dict['used']
-            ) for resource, r_dict in six.iteritems(self.resources())]
+            ) for resource, r_dict in self.resources().items()]
         if self._resources is not None:
             consumers = collections.defaultdict(dict)
             for task in self._state.get_active_tasks_by_status(RUNNING):
                 if task.status == RUNNING and task.resources:
-                    for resource, amount in six.iteritems(task.resources):
+                    for resource, amount in task.resources.items():
                         consumers[resource][task.id] = self._serialize_task(task.id, include_deps=False)
             for resource in resources:
                 tasks = consumers[resource['name']]
@@ -1504,7 +1524,7 @@ class Scheduler(object):
         ''' get total resources and available ones '''
         used_resources = self._used_resources()
         ret = collections.defaultdict(dict)
-        for resource, total in six.iteritems(self._resources):
+        for resource, total in self._resources.items():
             ret[resource]['total'] = total
             if resource in used_resources:
                 ret[resource]['used'] = used_resources[resource]
@@ -1541,7 +1561,9 @@ class Scheduler(object):
     def fetch_error(self, task_id, **kwargs):
         if self._state.has_task(task_id):
             task = self._state.get_task(task_id)
-            return {"taskId": task_id, "error": task.expl, 'displayName': task.pretty_id}
+            return {"taskId": task_id, "error": task.expl, 'displayName':
+                    task.pretty_id, 'taskParams': task.params, 'taskModule':
+                    task.module, 'taskFamily': task.family}
         else:
             return {"taskId": task_id, "error": ""}
 
@@ -1587,7 +1609,7 @@ class Scheduler(object):
                 return
 
             def decrease(resources, decrease_resources):
-                for resource, decrease_amount in six.iteritems(decrease_resources):
+                for resource, decrease_amount in decrease_resources.items():
                     if decrease_amount > 0 and resource in resources:
                         resources[resource] = max(0, resources[resource] - decrease_amount)
 
@@ -1620,3 +1642,7 @@ class Scheduler(object):
     def task_history(self):
         # Used by server.py to expose the calls
         return self._task_history
+
+    @rpc_method()
+    def update_metrics_task_started(self, task):
+        self._state._metrics_collector.handle_task_started(task)
