@@ -29,6 +29,7 @@ ways between versions. The exception is the exception types and the
 """
 
 import collections
+import datetime
 import getpass
 import importlib
 import logging
@@ -39,10 +40,7 @@ import subprocess
 import sys
 import contextlib
 
-try:
-    import Queue
-except ImportError:
-    import queue as Queue
+import queue as Queue
 import random
 import socket
 import threading
@@ -50,8 +48,6 @@ import time
 import traceback
 import types
 
-from luigi import six
-
 from luigi import notifications
 from luigi.event import Event
 from luigi.task_register import load_task
@@ -61,12 +57,9 @@ from luigi.target import Target
 from luigi.task import Task, flatten, getpaths, Config
 from luigi.task_register import TaskClassException
 from luigi.task_status import RUNNING
-from luigi.parameter import BoolParameter, FloatParameter, IntParameter, Parameter
+from luigi.parameter import BoolParameter, FloatParameter, IntParameter, OptionalParameter, Parameter, TimeDeltaParameter
 
-try:
-    import simplejson as json
-except ImportError:
-    import json
+import json
 
 logger = logging.getLogger('luigi-interface')
 
@@ -89,7 +82,7 @@ def _is_external(task):
 
 
 def _get_retry_policy_dict(task):
-    return RetryPolicy(task.retry_count, task.disable_hard_timeout, task.disable_window_seconds)._asdict()
+    return RetryPolicy(task.retry_count, task.disable_hard_timeout, task.disable_window)._asdict()
 
 
 class TaskException(Exception):
@@ -123,17 +116,18 @@ class TaskProcess(multiprocessing.Process):
     }
 
     def __init__(self, task, worker_id, result_queue, status_reporter,
-                 use_multiprocessing=False, worker_timeout=0, check_unfulfilled_deps=True):
+                 use_multiprocessing=False, worker_timeout=0, check_unfulfilled_deps=True,
+                 check_complete_on_run=False):
         super(TaskProcess, self).__init__()
         self.task = task
         self.worker_id = worker_id
         self.result_queue = result_queue
         self.status_reporter = status_reporter
-        if task.worker_timeout is not None:
-            worker_timeout = task.worker_timeout
-        self.timeout_time = time.time() + worker_timeout if worker_timeout else None
+        self.worker_timeout = task.worker_timeout if task.worker_timeout is not None else worker_timeout
+        self.timeout_time = time.time() + self.worker_timeout if self.worker_timeout else None
         self.use_multiprocessing = use_multiprocessing or self.timeout_time is not None
         self.check_unfulfilled_deps = check_unfulfilled_deps
+        self.check_complete_on_run = check_complete_on_run
 
     def _run_get_new_deps(self):
         task_gen = self.task.run()
@@ -145,7 +139,7 @@ class TaskProcess(multiprocessing.Process):
         while True:
             try:
                 if next_send is None:
-                    requires = six.next(task_gen)
+                    requires = next(task_gen)
                 else:
                     requires = task_gen.send(next_send)
             except StopIteration:
@@ -164,7 +158,9 @@ class TaskProcess(multiprocessing.Process):
 
         if self.use_multiprocessing:
             # Need to have different random seeds if running in separate processes
-            random.seed((os.getpid(), time.time()))
+            processID = os.getpid()
+            currentTime = time.time()
+            random.seed((processID, currentTime))
 
         status = FAILED
         expl = ''
@@ -186,8 +182,6 @@ class TaskProcess(multiprocessing.Process):
 
             if _is_external(self.task):
                 # External task
-                # TODO(erikbern): We should check for task completeness after non-external tasks too!
-                # This will resolve #814 and make things a lot more consistent
                 if self.task.complete():
                     status = DONE
                 else:
@@ -197,7 +191,13 @@ class TaskProcess(multiprocessing.Process):
             else:
                 with self._forward_attributes():
                     new_deps = self._run_get_new_deps()
-                status = DONE if not new_deps else PENDING
+                if not new_deps:
+                    if not self.check_complete_on_run or self.task.complete():
+                        status = DONE
+                    else:
+                        raise TaskException("Task finished running, but complete() is still returning false.")
+                else:
+                    status = PENDING
 
             if new_deps:
                 logger.info(
@@ -215,15 +215,17 @@ class TaskProcess(multiprocessing.Process):
             raise
         except BaseException as ex:
             status = FAILED
-            logger.exception("[pid %s] Worker %s failed    %s", os.getpid(), self.worker_id, self.task)
-            self.task.trigger_event(Event.FAILURE, self.task, ex)
-            raw_error_message = self.task.on_failure(ex)
-            expl = raw_error_message
+            expl = self._handle_run_exception(ex)
 
         finally:
             self.result_queue.put(
                 (self.task.task_id, status, expl, missing, new_deps))
 
+    def _handle_run_exception(self, ex):
+        logger.exception("[pid %s] Worker %s failed    %s", os.getpid(), self.worker_id, self.task)
+        self.task.trigger_event(Event.FAILURE, self.task, ex)
+        return self.task.on_failure(ex)
+
     def _recursive_terminate(self):
         import psutil
 
@@ -255,13 +257,13 @@ class TaskProcess(multiprocessing.Process):
     @contextlib.contextmanager
     def _forward_attributes(self):
         # forward configured attributes to the task
-        for reporter_attr, task_attr in six.iteritems(self.forward_reporter_attributes):
+        for reporter_attr, task_attr in self.forward_reporter_attributes.items():
             setattr(self.task, task_attr, getattr(self.status_reporter, reporter_attr))
         try:
             yield self
         finally:
             # reset attributes again
-            for reporter_attr, task_attr in six.iteritems(self.forward_reporter_attributes):
+            for reporter_attr, task_attr in self.forward_reporter_attributes.items():
                 setattr(self.task, task_attr, None)
 
 
@@ -285,7 +287,7 @@ class ContextManagedTaskProcess(TaskProcess):
             super(ContextManagedTaskProcess, self).run()
 
 
-class TaskStatusReporter(object):
+class TaskStatusReporter:
     """
     Reports task status information to the scheduler.
 
@@ -316,7 +318,7 @@ class TaskStatusReporter(object):
         self._scheduler.decrease_running_task_resources(self._task_id, decrease_resources)
 
 
-class SchedulerMessage(object):
+class SchedulerMessage:
     """
     Message object that is build by the the :py:class:`Worker` when a message from the scheduler is
     received and passed to the message queue of a :py:class:`Task`.
@@ -342,7 +344,7 @@ class SchedulerMessage(object):
         self._scheduler.add_scheduler_message_response(self._task_id, self._message_id, response)
 
 
-class SingleProcessPool(object):
+class SingleProcessPool:
     """
     Dummy process pool for using a single processor.
 
@@ -383,7 +385,7 @@ class AsyncCompletionException(Exception):
         self.trace = trace
 
 
-class TracebackWrapper(object):
+class TracebackWrapper:
     """
     Class to wrap tracebacks so we can know they're not just strings.
     """
@@ -407,6 +409,8 @@ def check_complete(task, out_queue):
 class worker(Config):
     # NOTE: `section.config-variable` in the config_path argument is deprecated in favor of `worker.config_variable`
 
+    id = Parameter(default='',
+                   description='Override the auto-generated worker_id')
     ping_interval = FloatParameter(default=1.0,
                                    config_path=dict(section='core', name='worker-ping-interval'))
     keep_alive = BoolParameter(default=False,
@@ -424,6 +428,8 @@ class worker(Config):
                                    config_path=dict(section='core', name='worker-wait-interval'))
     wait_jitter = FloatParameter(default=5.0)
 
+    max_keep_alive_idle_duration = TimeDeltaParameter(default=datetime.timedelta(0))
+
     max_reschedules = IntParameter(default=1,
                                    config_path=dict(section='core', name='worker-max-reschedules'))
     timeout = IntParameter(default=0,
@@ -443,15 +449,19 @@ class worker(Config):
     check_unfulfilled_deps = BoolParameter(default=True,
                                            description='If true, check for completeness of '
                                            'dependencies before running a task')
+    check_complete_on_run = BoolParameter(default=False,
+                                          description='If true, only mark tasks as done after running if they are complete. '
+                                          'Regardless of this setting, the worker will always check if external '
+                                          'tasks are complete before marking them as done.')
     force_multiprocessing = BoolParameter(default=False,
                                           description='If true, use multiprocessing also when '
                                           'running with 1 worker')
-    task_process_context = Parameter(default=None,
-                                     description='If set to a fully qualified class name, the class will '
-                                     'be instantiated with a TaskProcess as its constructor parameter and '
-                                     'applied as a context manager around its run() call, so this can be '
-                                     'used for obtaining high level customizable monitoring or logging of '
-                                     'each individual Task run.')
+    task_process_context = OptionalParameter(default=None,
+                                             description='If set to a fully qualified class name, the class will '
+                                             'be instantiated with a TaskProcess as its constructor parameter and '
+                                             'applied as a context manager around its run() call, so this can be '
+                                             'used for obtaining high level customizable monitoring or logging of '
+                                             'each individual Task run.')
 
 
 class KeepAliveThread(threading.Thread):
@@ -494,7 +504,7 @@ def rpc_message_callback(fn):
     return fn
 
 
-class Worker(object):
+class Worker:
     """
     Worker object communicates with a scheduler.
 
@@ -511,11 +521,10 @@ class Worker(object):
         self.worker_processes = int(worker_processes)
         self._worker_info = self._generate_worker_info()
 
-        if not worker_id:
-            worker_id = 'Worker(%s)' % ', '.join(['%s=%s' % (k, v) for k, v in self._worker_info])
-
         self._config = worker(**kwargs)
 
+        worker_id = worker_id or self._config.id or self._generate_worker_id(self._worker_info)
+
         assert self._config.wait_interval >= _WAIT_INTERVAL_EPS, "[worker] wait_interval must be positive"
         assert self._config.wait_jitter >= 0.0, "[worker] wait_jitter must be equal or greater than zero"
 
@@ -549,6 +558,7 @@ class Worker(object):
         # Keep info about what tasks are running (could be in other processes)
         self._task_result_queue = multiprocessing.Queue()
         self._running_tasks = {}
+        self._idle_since = None
 
         # Stuff for execution_summary
         self._add_task_history = []
@@ -598,6 +608,7 @@ class Worker(object):
         for task in self._running_tasks.values():
             if task.is_alive():
                 task.terminate()
+        self._task_result_queue.close()
         return False  # Don't suppress exception
 
     def _generate_worker_info(self):
@@ -625,6 +636,10 @@ class Worker(object):
             pass
         return args
 
+    def _generate_worker_id(self, worker_info):
+        worker_info_str = ', '.join(['{}={}'.format(k, v) for k, v in worker_info])
+        return 'Worker({})'.format(worker_info_str)
+
     def _validate_task(self, task):
         if not isinstance(task, Task):
             raise TaskException('Can not schedule non-task %s' % task)
@@ -1008,6 +1023,7 @@ class Worker(object):
             use_multiprocessing=use_multiprocessing,
             worker_timeout=self._config.timeout,
             check_unfulfilled_deps=self._config.check_unfulfilled_deps,
+            check_complete_on_run=self._config.check_complete_on_run,
         )
 
     def _purge_children(self):
@@ -1016,13 +1032,13 @@ class Worker(object):
 
         :return:
         """
-        for task_id, p in six.iteritems(self._running_tasks):
+        for task_id, p in self._running_tasks.items():
             if not p.is_alive() and p.exitcode:
                 error_msg = 'Task {} died unexpectedly with exit code {}'.format(task_id, p.exitcode)
                 p.task.trigger_event(Event.PROCESS_FAILURE, p.task, error_msg)
             elif p.timeout_time is not None and time.time() > float(p.timeout_time) and p.is_alive():
                 p.terminate()
-                error_msg = 'Task {} timed out after {} seconds and was terminated.'.format(task_id, p.task.worker_timeout)
+                error_msg = 'Task {} timed out after {} seconds and was terminated.'.format(task_id, p.worker_timeout)
                 p.task.trigger_event(Event.TIMEOUT, p.task, error_msg)
             else:
                 continue
@@ -1039,6 +1055,7 @@ class Worker(object):
            will be rescheduled and dependencies added,
         3. child process dies: we need to catch this separately.
         """
+        self._idle_since = None
         while True:
             self._purge_children()  # Deal with subprocess failures
 
@@ -1127,8 +1144,16 @@ class Worker(object):
             return get_work_response.n_pending_last_scheduled > 0
         elif self._config.count_uniques:
             return get_work_response.n_unique_pending > 0
+        elif get_work_response.n_pending_tasks == 0:
+            return False
+        elif not self._config.max_keep_alive_idle_duration:
+            return True
+        elif not self._idle_since:
+            return True
         else:
-            return get_work_response.n_pending_tasks > 0
+            time_to_shutdown = self._idle_since + self._config.max_keep_alive_idle_duration - datetime.datetime.now()
+            logger.debug("[%s] %s until shutdown", self._id, time_to_shutdown)
+            return time_to_shutdown > datetime.timedelta(0)
 
     def handle_interrupt(self, signum, _):
         """
@@ -1170,8 +1195,9 @@ class Worker(object):
                 if not self._stop_requesting_work:
                     self._log_remote_tasks(get_work_response)
                 if len(self._running_tasks) == 0:
+                    self._idle_since = self._idle_since or datetime.datetime.now()
                     if self._keep_alive(get_work_response):
-                        six.next(sleeper)
+                        next(sleeper)
                         continue
                     else:
                         break
