@@ -18,7 +18,7 @@ import doctest
 import pickle
 import warnings
 
-from helpers import unittest, LuigiTestCase
+from helpers import unittest, LuigiTestCase, with_config
 from datetime import datetime, timedelta
 
 import luigi
@@ -152,6 +152,72 @@ class TaskTest(unittest.TestCase):
         with self.assertWarnsRegex(UserWarning, 'Parameter "timedelta_param" with value ".*" is not of type timedelta.'):
             DummyTask(**params)
 
+    def test_disable_window_seconds(self):
+        """
+        Deprecated disable_window_seconds param uses disable_window value
+        """
+        class ATask(luigi.Task):
+            disable_window = 17
+        task = ATask()
+        self.assertEqual(task.disable_window_seconds, 17)
+
+    @with_config({"ATaskWithBadParam": {"bad_param": "bad_value"}})
+    def test_bad_param(self):
+        class ATaskWithBadParam(luigi.Task):
+            bad_param = luigi.IntParameter()
+
+        with self.assertRaisesRegex(ValueError, r"ATaskWithBadParam\[args=\(\), kwargs={}\]: Error when parsing the default value of 'bad_param'"):
+            ATaskWithBadParam()
+
+    @with_config(
+        {
+            "TaskA": {
+                "a": "a",
+                "b": "b",
+                "c": "c",
+            },
+            "TaskB": {
+                "a": "a",
+                "b": "b",
+                "c": "c",
+            },
+        }
+    )
+    def test_unconsumed_params(self):
+        class TaskA(luigi.Task):
+            a = luigi.Parameter(default="a")
+
+        class TaskB(luigi.Task):
+            a = luigi.Parameter(default="a")
+
+        with warnings.catch_warnings(record=True) as w:
+            warnings.filterwarnings(
+                action="ignore",
+                category=Warning,
+            )
+            warnings.simplefilter(
+                action="always",
+                category=luigi.parameter.UnconsumedParameterWarning,
+            )
+
+            TaskA()
+            TaskB()
+
+            assert len(w) == 4
+            expected = [
+                ("b", "TaskA"),
+                ("c", "TaskA"),
+                ("b", "TaskB"),
+                ("c", "TaskB"),
+            ]
+            for i, (expected_value, task_name) in zip(w, expected):
+                assert issubclass(i.category, luigi.parameter.UnconsumedParameterWarning)
+                assert str(i.message) == (
+                    "The configuration contains the parameter "
+                    f"'{expected_value}' with value '{expected_value}' that is not consumed by "
+                    f"the task '{task_name}'."
+                )
+
 
 class ExternalizeTaskTest(LuigiTestCase):
 
@@ -392,3 +458,15 @@ class AutoNamespaceTest(LuigiTestCase):
             pass
         luigi.namespace(scope='incorrect_namespace')
         self.assertEqual(MyTask.get_task_namespace(), '')
+
+
+class InitSubclassTest(LuigiTestCase):
+    def test_task_works_with_init_subclass(self):
+        class ReceivesClassKwargs(luigi.Task):
+            def __init_subclass__(cls, x, **kwargs):
+                super(ReceivesClassKwargs, cls).__init_subclass__()
+                cls.x = x
+
+        class Receiver(ReceivesClassKwargs, x=1):
+            pass
+        self.assertEquals(Receiver.x, 1)
